from fastapi import APIRouter, HTTPException, File, UploadFile
from fRAGme.util.v1.chroma_handler import *

router = APIRouter()


@router.post("/add_texts")
def data_add_text(texts: List[Text], identifier: str):
    try:
        add_texts(texts, identifier)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

    return {"result": True}


@router.post("/add_pdfs")
def data_add_pdfs(identifier: str, pdfs: List[UploadFile] = File(...)):
    try:
        add_pdfs(identifier, pdfs)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

    return {"result": True}


@router.post("/get_texts")
def data_get_texts(identifier: str, ids: List[str]):
    try:
        documents = get_texts(identifier, ids)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))
    return {"result": documents}


@router.get("/get_pdfs")
def data_get_pdfs(identifier: str):
    try:
        documents = get_pdfs(identifier)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))
    return {"result": documents}


@router.get("/get_databases")
def data_get_databases():
    try:
        documents = get_databases()
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))
    return {"result": documents}


@router.put("/update_texts")
def data_update_texts(identifier: str, updates: Dict[str, TextUpdate]):
    try:
        update_texts(identifier, updates)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))
    return {"result": True}


@router.delete("/delete_texts")
def data_delete_texts(identifier: str, ids: List[str]):
    try:
        delete_texts(identifier, ids)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))
    return {"result": True}


@router.delete("/delete_pdfs")
def data_delete_pdfs(identifier: str, pdf_names: List[str]):
    try:
        delete_pdfs(identifier, pdf_names)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))
    return {"result": True}


@router.delete("/delete_databases")
def data_delete_databases(identifiers: List[str]):
    try:
        delete_databases(identifiers)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))
    return {"result": True}



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

from fRAGme import __version__ as fRAGme_version

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "fRAGme"
copyright = "2024, Henning Krause"
author = "Henning Krause"

release = fRAGme_version
version = fRAGme_version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinxcontrib.spelling",
    "sphinx_sitemap",
    "matplotlib.sphinxext.plot_directive",
    "myst_parser",
    "nbsphinx",
    "nbsphinx_link",
    "jupyter_sphinx",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

autodoc_default_options = {"members": True, "inherited-members": True}


# Skip property members --> They should be defined in Attributes
def skip_property_member(app, what, name, obj, skip, options):
    if isinstance(obj, property):
        return True


def setup(app):
    app.connect("autodoc-skip-member", skip_property_member)


templates_path = ["_templates"]
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "README.md",
    "**.ipynb_checkpoints",
]

autosummary_generate = True

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_logo = "_static/logo.png"
html_favicon = "_static/favicon.png"
html_theme_options = {
    "logo_only": True,
    "display_version": True,
}
html_css_files = [
    "css/custom.css",
]

latex_engine = "xelatex"
latex_elements = {
    "preamble": r"""
    \usepackage{braket}
    """,
}

# intersphinx
intersphinx_mapping = {
    "qiskit": ("https://docs.quantum.ibm.com/api/qiskit/", None),
    "qiskit-aer": ("https://qiskit.github.io/qiskit-aer/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
}

suppress_warnings = ["myst.header", "config.cache"]

# base URL for sphinx_sitemap
html_baseurl = "https://krauhen.github.io/fRAGme/"
sitemap_url_scheme = "{link}"



from fastapi import APIRouter, HTTPException
from fRAGme.util.v1.chroma_handler import *
from fRAGme.models.v1.cmd import *
from openai import OpenAI

router = APIRouter()


@router.post("/ask_question")
def cmd_ask_question(request: Question, identifier: str):
    try:
        prompt = build_question(request, identifier)
        client = OpenAI()

        chat_history = [
            {"role": element.role, "content": element.content}
            for element in request.chat_history
        ]
        chat_history.append({"role": RoleEnum.user, "content": prompt})

        completion = client.chat.completions.create(
            model="gpt-4o-mini", messages=chat_history
        )

        answer = completion.choices[0].message
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

    return {"result": answer}



import os

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fRAGme.api.v1.data import router as data_router
from fRAGme.api.v1.cmd import router as cmd_router
from fRAGme.util.v1.chroma_handler import get_vector_store
from dotenv import load_dotenv


load_dotenv()

app = FastAPI(
    title="fRAGme",
    description="Retrieval Augmented Generation (RAG) Service.",
)
app.include_router(data_router, prefix="/data/v1", tags=["data"])
app.include_router(cmd_router, prefix="/cmd/v1", tags=["cmd"])

app.add_middleware(
    CORSMiddleware,
    allow_origins=[os.getenv("ORIGIN", "*")],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.on_event("startup")
def start_app():
    get_vector_store("base")


@app.on_event("shutdown")
def stop_app():
    pass


@app.get("/")
def healthcheck():
    return True



import glob
import os
import shutil

from typing import List
from uuid import uuid4
from langchain_openai import OpenAIEmbeddings
from langchain_chroma import Chroma
from langchain_core.documents import Document
from langchain_community.document_loaders import PyPDFLoader
from fRAGme.models.v1.data import *
from fRAGme.models.v1.cmd import *

vector_stores = dict()


def create_vector_store(identifier):
    data_path = os.getenv("DATA_PATH")
    filename = os.path.join(data_path, f"{identifier}_chroma_langchain_db")
    embeddings = OpenAIEmbeddings(model="text-embedding-3-large")
    vector_store = Chroma(
        collection_name=identifier,
        embedding_function=embeddings,
        persist_directory=filename,
    )
    return vector_store


def get_vector_store(identifier):
    if identifier not in vector_stores.keys() or vector_stores[identifier] is None:
        vector_store = create_vector_store(identifier)
        vector_stores[identifier] = vector_store
    else:
        vector_store = vector_stores[identifier]
    return vector_store


def add_texts(texts: List[Text], identifier: str):
    documents = list()
    for element in texts:
        document = Document(
            page_content=element.text,
            metadata=element.metadata,
        )
        documents.append(document)
    uuids = [str(uuid4()) for _ in range(len(documents))]

    vector_store = get_vector_store(identifier)
    vector_store.add_documents(documents=documents, ids=uuids)


def add_pdfs(identifier: str, pdfs):
    documents = list()
    for pdf in pdfs:
        file_path = os.path.join("/tmp", pdf.filename)
        with open(file_path, "wb") as f:
            f.write(pdf.file.read())

        loader = PyPDFLoader(file_path)
        pages = loader.load_and_split()

        os.remove(file_path)

        documents.extend(pages)

    if len(documents) > 0:
        uuids = [str(uuid4()) for _ in range(len(documents))]

        vector_store = get_vector_store(identifier)
        vector_store.add_documents(documents=documents, ids=uuids)
    else:
        raise Exception("PDF has no text pages.(Maybe all pages are images?!)")


def get_texts(identifier: str, ids: List[str] = None):
    vector_store = get_vector_store(identifier)
    if ids is not None:
        documents = vector_store.get(ids=ids)
    else:
        documents = vector_store.get()

    elements = dict()
    for id_, text, metadata in zip(
        documents["ids"], documents["documents"], documents["metadatas"]
    ):
        elements[id_] = dict()
        elements[id_]["text"] = text
        elements[id_]["metadata"] = metadata

    return elements


def get_pdfs(identifier: str):
    documents = get_texts(identifier)
    filenames = list()
    for id_, element in documents.items():
        if "source" in element["metadata"].keys():
            source = element["metadata"]["source"]
            filename = source.split("/")[-1]
            extension = filename.split(".")[-1]
            if extension == "pdf":
                filenames.append(filename)
    filenames = list(set(filenames))
    return filenames


def get_databases():
    data_path = os.getenv("DATA_PATH")
    databases = glob.glob(os.path.join(data_path, "*_chroma_langchain_db"))
    databases = [
        database.split("/")[-1].replace("_chroma_langchain_db", "")
        for database in databases
    ]
    return databases


def update_texts(identifier: str, updates: Dict[str, TextUpdate]):
    vector_store = get_vector_store(identifier)
    ids = list(updates.keys())
    documents = vector_store.get(ids=ids)
    for id_, metadata in zip(documents["ids"], documents["metadatas"]):
        document = Document(
            page_content=updates[id_].new_text,
            metadata=updates[id_].new_metadata,
        )
        vector_store.update_document(id_, document)


def delete_texts(identifier: str, ids: List[str]):
    vector_store = get_vector_store(identifier)
    vector_store.delete(ids)


def delete_pdfs(identifier: str, pdf_names: List[str]):
    documents = get_texts(identifier)
    ids = list()
    for pdf_name in pdf_names:
        for id_, element in documents.items():
            if "source" in element["metadata"].keys():
                source = element["metadata"]["source"]
                filename = source.split("/")[-1]
                if pdf_name in filename:
                    ids.append(id_)
    ids = list(set(ids))
    if len(ids) > 0:
        delete_texts(identifier, ids)


def delete_databases(identifiers: List[str]):
    for identifier in identifiers:
        data_path = os.getenv("DATA_PATH")
        filepath = os.path.join(data_path, f"{identifier}_chroma_langchain_db")
        if os.path.isdir(filepath):
            vector_stores[identifier] = None
            shutil.rmtree(filepath)


def build_question(data: Question, identifier: str):
    # Add question
    template = f"Question:\n{data.question}\n\n"

    # Add k snippets
    template += f"Info-Snippets:\n"
    vector_store = get_vector_store(identifier)
    snippets = vector_store.similarity_search(
        data.question, k=data.k_similar_text_snippets
    )
    for snippet in snippets:
        template += f"Text: {snippet.page_content}\n"
        template += f"Metadata: {snippet.metadata}"
        template += f"\n\n"

    return template



from enum import Enum
from typing import List
from pydantic import BaseModel

BASE_PROMPT = (
    "You are a RAG assistant and should answer questions according to the Info-Snippet you get.\n"
    "The prompt will have the following syntax:\n"
    "Question:\n{data.question}\n\n"
    "Info-Snippets:\n"
    "Text: {snippet.page_content}\n"
    "Metadata: {snippet.metadata}"
    "\n\n"
    "Text: {snippet.page_content}\n"
    "Metadata: {snippet.metadata}"
    "\n\n"
    "Text: {snippet.page_content}\n"
    "Metadata: {snippet.metadata}"
    "\n\n"
    "..."
)


class RoleEnum(str, Enum):
    user = "user"
    system = "system"
    assistant = "assistant"


class ChatAction(BaseModel):
    role: RoleEnum
    content: str


class Question(BaseModel):
    base_prompt: str = BASE_PROMPT
    chat_history: List[ChatAction]
    question: str
    k_similar_text_snippets: int = 2



from typing import Dict, Any
from pydantic import BaseModel


class Text(BaseModel):
    text: str
    metadata: Dict[str, Any] = {"source": "text input"}


class TextUpdate(BaseModel):
    new_text: str
    new_metadata: Dict[str, Any]
